"""
argument-aliaser

   The purpose of salias is to provide an aliasing system for programs
   where arguments may wish to be aliased. This is more useful than just
   a shell environment variable as an argument, since it can 'repeat'
   options, such as GnuPG's --recipient option as needed. Also, unlike
   shell enviornment variables, it can be used transparently inside of
   other programs, strengthening it's drop-in capabilities.

   salias was initially designed for wrapping around GnuPG, providing
   aliases for keys. However, the design easily extends to being useful
   for all types of programs.

Usage

   Simply call salias before the the program it is wrapping, similar to
   how you do for time(1). For example: salias gpg -r mykey --encrypt

   Another option is to have a wrapper script act as a fully transparent
   drop-in wrapper. See the section on Transparent Wrapping for details.

Options

   --context name
          Use name as the context name instead of the program that salias
          is given.

   --config file
          Use file for configuration, instead of the default
          ~/.salias/config.xml

   --debug
          Instead of running the specified program, merely print out what
          would be executed, with the aliased arguments expanded.

   --quiet
          Surpress any warning (non-fatal) errors.

Configuration

   The configuration file must be a well-formed XML document. Instead of
   starting you off with a DTD or Schema, it'd probably just be best to
   start off with an example that shows many of the possible things that
   can be configured. Here's a verbose example of a configuration for
   GnuPG (gpg) (with lots of whitespace for clarity):

<config>
  <context name="gpg">

    <aliases>

      <alias name="frank">
        <value>4F863BBBA8166F0A340F600356FFD10A260C4FA3</value>
      </alias>

     <alias name="mykey">
       <value>0x1234ABCD</value>
     </alias>

     <alias name="project-list">
       <value>frank</value>
       <value>bob</value>
       <value>mykey</value>
     </alias>

   </aliases>

   <repeated-options>
     <option>-r</option>
     <option>--recipient</option>
   </repeated-options>

  </context>
</config>

   You don't have to have specify all of these elements; the more
   advanced ones such repeated-options can be set with sensible defaults
   for some programs by setting the default attribute of context element.
   Defaults will fill in the repeated-options part of the context
   element, so you don't have to. Using defaults, the above example could
   be shortened to:

<config>
  <context name="gpg">

    <aliases>

      <alias name="frank">
        <value>4F863BBBA8166F0A340F600356FFD10A260C4FA3</value>
      </alias>

     <alias name="mykey">
       <value>0x1234ABCD</value>
     </alias>

     <alias name="project-list">
       <value>frank</value>
       <value>bob</value>
       <value>mykey</value>
     </alias>

   </aliases>

  </context>
</config>

   By default, the program which salias is wrapping around is used to
   determine the context to use. However, if --context is specified as an
   argument to salias, this behaviour is overriden, and the option's
   value is used as the context to use.

  Configuration elements

   config
          Each configuration contain a single config, which encompasses
          the setup. config currently contains multiple context elements.
          No attributes.

   context
          Embodies how to process and alias arguments when dealing with
          programs of the name name. Has a mandatory name attribute, and
          an optional default attribute; see the section Defaults for
          more information. It also has optional child-elements of
          repeated-options and aliases.

   aliases
          Simply holds the list of alias elements. No attributes.

   alias
          Holds the information for an alias. Must have an attribute name
          which is the alias's name. Contains zero or more value
          elements.

   value
          Is a value for an alias; that is, it is the substituted key
          information for an alias. This can be the name of another
          alias, and it would be handled recursively. However, salias
          will not detect if you have infinitely recursive aliases, and
          will likely fly into an infinite loop if you have them.

   repeated-options
          Contains a list of option elements. These options are those
          which would need to be repeated if there are multiple
          substitions. For example, in GnuPG, you need to repeatedly put
          -r or --recipient before each encryption recipient. Hence,
          these would be repeated-options. On the other hand, you don't
          want to repeat --list-keys before every key, so this would not
          be in the list for GnuPG.

   option
          A program option, such as --recipient (including applicable
          dashes).

Defaults

   context elements elements can have an attribute of default, which can
   set some default, useful values for the repeated-optionsvalues.
   Currently supported defaults are:

   gnupg
          This default will set repeated-options to have options of -r
          and --recipient.

Examples

   Assuming salias has been configured to wrap around GnuPG (gpg), and
   the user has set the appropriate aliases as in the example config.xml
   above, here are some example argument transformations:

   salias gpg -r frank -r mykey --encrypt
          gpg -r 4F863BBBA8166F0A340F600356FFD10A260C4FA3 -r 0x1234ABCD
          --encrypt

   salias gpg --list-keys frank mykey
          gpg --list-keys 4F863BBBA8166F0A340F600356FFD10A260C4FA3
          0x1234ABCD

   salias gpg -r project-list --encrypt
          gpg -r 4F863BBBA8166F0A340F600356FFD10A260C4FA3 -r bob -r
          0x1234ABCD --encrypt

   salias gpg --default-key mykey --sign
          gpg --default-key 0x1234ABCD --sign

Transparent Wrapping

   Sure, it's nice to be able to alias arguments when using the shell and
   prefixing the command salias, but what about other programs that you
   might want to use these aliases transparently? You don't want to
   modify them to have to call salias, or have any knowledge of it
   whatsoever. What you want is a completely transparent drop-in.

   There are two ways to go about this: you can simply alias, say, gpg to
   salias gpg, or, so that other programs which would use gpg directly,
   you can create a wrapper script.

  Using a wrapper script

   Fortunately, you can insert salias transparently by using a wrapper
   script around the calls to the real program you are aliasing for. Take
   for instance, gpg. By having the file named /usr/local/bin/gpg
   actually be a wrapper script that calls /usr/local/bin.real/gpg, we
   can insert salias transparently.

   Here's an example of a flexible wrapper script that you could link
   every executable you want salias to handle on the system to, after
   renaming the real programs to /usr/local/bin.real/:

#!/bin/sh
base=`basename $0`
exec salias --context $base /usr/local/bin.real/$base "$@"

   For example, you could simply move the real gpg to
   /usr/local/bin.real/gpg, link /usr/local/bin/gpg to this wrapper
   script, and it would call salias, which in turn would call the real
   gpg properly, with the aliased arguments.

Files

   ~/.salias/config.xmlx
          Configuration file. If --config is used, use it instead.

Author

   Frank J. Tobin, ftobin@neverending.org

   OpenPGP key fingerprint:
4F86 3BBB A816 6F0A 340F  6003 56FF D10A 260C 4FA3

Copyright

   Copyright (C) 2001 Frank J. Tobin, ftobin@neverending.org

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of the
   License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA or see http://www.gnu.org/copyleft/lesser.html
"""

__author__   = "Frank J. Tobin, ftobin@neverending.org"
__version__  = "0.2.1"
__revision__ = "$Id: salias.py,v 1.8 2002/04/02 00:21:21 ftobin Exp $"

import os
import os.path
import sys
import getopt
import xml.sax

config_fn = 'config.xml'
subhomedir = '.salias'

class UsageError(Exception):
    pass

class NoContextError(Exception):
    pass

class Context(object):
    __slots__ = ['name', 'default', 'aliases', 'repeated_options']
    defaults = {'gnupg': ['-r', '--recipient']}
    
    def __init__(self, name, default=None):
        self.name = name
        self.default = default
        self.aliases = {}
        self.repeated_options = []
        if default != None:
            if self.defaults.has_key(default):
                self.repeated_options = self.defaults[default]
            else:
                sys.sterr.write("unknown default '%s'" % default)

    def add_alias(self, name):
        self.aliases[name] = []

    def add_alias_value(self, name, value):
        self.aliases[name].append(value)

    def add_repeated_option(self, name):
        self.repeated_options.append(name)

    def replace(self, oldargs):
        args = oldargs[:]
        # we deal with what we're processing in reverse order,
        # so that we can treat it more like a stack
        args.reverse()
        newargs = []
        repeat_prev = None

        while len(args) > 0:
            a = args.pop()
            subs = self.aliases.get(a, None)
            if subs is None:
                if a in self.repeated_options:
                    repeat_prev = a
                else:
                    repeat_prev = None
                newargs.append(a)
            else:
                # remember, keeping our incoming arglist in reverse...
                subs.reverse()
                # We need to erase our previous repeated option
                # because it has already been added to the argument
                # list earlier in the processing, but we'll
                # re-add it here
                if repeat_prev != None:
                    newargs.pop()

                for s in subs:
                    args.append(s)
                    if repeat_prev != None:
                        args.append(repeat_prev)
        return newargs


class ConfigHandler(xml.sax.ContentHandler):
    _char_els = ['value', 'option']
    _parents  = {'config': None,
                 'context': 'config',
                 'aliases': 'context',
                 'alias':   'aliases',
                 'value':   'alias',
                 'repeated-options': 'context',
                 'option': 'repeated-options',
                 }
    
    _handlers = ['startElement', 'endElement', 'characters']

    def __init__(self, search_for):
        self._search_for = search_for
    
    def startDocument(self):
        self._skip = 0
        self._cxt = None
        self._path = []
        self._cur_alias_name = None
        self._cur_chars = None

    def _check_valid(self, name):
        """Ensure the document is kinda valid"""
        if self._parents.get(name, -1) != self._parent():
            raise xml.sax.SAXParseException(("element '%s' not a valid child element of '%s'"
                                             % (name, self._parent())),
                                            None, self._locator)

    def startElement(self, name, attrs):
        self._check_valid(name)

        # we turn this off if we don't find a 'match'
        if not self._skip:
            if name == 'context' and self._search_for != attrs['name']:
                self._skip = 1
            else:
                if name == 'context':
                    cxt = Context(attrs['name'],
                                  attrs.get('default', None))
                    self._cxt = cxt
                elif name == 'alias':
                    self._cxt.add_alias(attrs['name'])
                    self._cur_alias_name = attrs['name']
                    
                elif name in self._char_els:
                    self._cur_chars = ''
        
        self._path.append(name)


    def endElement(self, name):        
        if name == 'context':
            if self._skip:
                self._skip = 0
            else:
                # found it, I'm done!
                self._make_impotent()

        if not self._skip:
            if name == 'value':
                self._cxt.add_alias_value(self._cur_alias_name,
                                          self._cur_chars)
            elif name == 'option':
                self._cxt.add_repeated_option(self._cur_chars)
                
                if name in self._char_els:
                    self._cur_chars = None
        
        self._path.pop()


    def characters(self, content):
        if self._cur_chars != None:
            self._cur_chars += content

    def _passthrough(*args):
        pass

    def _make_impotent(self):
        for h in self._handlers:
            setattr(self, h, self._passthrough)

    def _parent(self):
        if len(self._path) == 0: return None
        return self._path[-1]

    def get_cxt(self):
        if self._cxt: return self._cxt
        raise NoContextError, self._search_for


class Options(object):
    __slots__ = ['_config', 'debug', 'quiet']

    def __init__(self):
        self.debug = 0
        self._config = None
        self.quiet = 0


class Salias(object):
    __slots__ = ['debug', 'quiet', 'config']

    def __init__(self, debug=0, quiet=0, config=None):
        self.debug = debug
        self.quiet = quiet
        self.config = config

    def run(self, argv=sys.argv):
        """called as salias is from the command line.
        This will exec if debug is not on."""
        new_args = None
        try:
            new_args = self.get_new_args(argv[1:])
        except UsageError, e:
            usage(e)
        except getopt.GetoptError, e:
            usage(e)

        if self.debug:
            print ' '.join(map(shell_quote, new_args))
        else:
            os.execvp(new_args[0], new_args)


    def get_new_args(self, argv, context=None):
        """context will be determined from argv if context==None.
        argv should NOT contain the program name, but only its args,
        unless context is None."""
        if context == None:
            (context, argv) = self.update_from_mixed_args(argv)
            if len(argv) == 0:
                raise UsageError, "no program given"
            if context == None:
                context = os.path.basename(argv[0])

        config = self.config
        if config == None:
            config = get_default_config()

        if config == None:
            #self._warn("configuration file unavailable (not specified and default location undetermined); proceeding")
            return argv
        
        if not os.path.exists(config):
            #self._warn("'%s' does not exist; proceeding" % config)
            return argv
        try:
            cxt = self.get_cxt(context, config)
        except NoContextError, e:
            #self._warn("No context '%s' in %s; proceeding\n"
            #           % (context, config))
            return argv
        
        return cxt.replace(argv)

    def get_cxt(self, cxt_name, config):
        parser = xml.sax.make_parser()
        handler = ConfigHandler(cxt_name)
        parser.setContentHandler(handler)
        parser.parse(config)
        return handler.get_cxt()
        

    def update_from_mixed_args(self, argv):
        """determine several instance attributes from
        argv which is a list of arguments of both salias
        and program arguments"""
        (opts, rest) = getopt.getopt(argv, '-q',
                                     ['config=', 'context=',
                                      'debug', 'quiet'])
        context = None
        
        for (o, v) in opts:
            if o == '--config':
                self.config = v
            elif o == '--context':
                context = v
            elif o == '--debug':
                self.debug = 1
            elif o == '-q' or o == '--quiet':
                self.quiet = 1

        return (context, rest)


    def _warn(self, s):
        if not self.quiet:
            sys.stderr.write('%s\n' % s)


def usage(e=None):
    """print out usage and exit.  e is an optional error message"""
    if e != None:
        sys.stderr.write('%s\n' % e)
    sys.stderr.write('usage: %s [options] program [args]\n' % sys.argv[0])
    sys.exit(1)


def get_default_config():
    userhome = os.getenv('HOME')
    if userhome == None: return None
    return os.path.join(userhome, subhomedir, config_fn)

def shell_quote(s):
    """return s as it would look in the shell as one string
    (e.g., escape/quote it properly)
    Tips grabbed from:
    http://www.mpi-sb.mpg.de/~uwe/lehre/unixffb/quoting-guide.html
    """
    return """'%s'""" % s.replace("""'""", """'\\''""")
